/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::borrow::Cow;
use std::rc::Rc;

use html5ever::driver::parse_fragment;
use html5ever::interface::QualName;
use html5ever::tendril::TendrilSink;
use html5ever::{local_name, namespace_url, ns};
use html_escape::encode_text_minimal;
use markup5ever_rcdom::{Handle, Node, NodeData, RcDom};
use ratatui::style::{Style, Stylize};
use ratatui::text::{Line, Span, Text};

pub fn escape_html<S: ?Sized + AsRef<str>>(s: &S) -> Cow<'_, str> {
  encode_text_minimal(s)
}

fn count_digits(n: u16) -> u8 {
  f32::log10(n.into()).floor() as u8 + 1
}

#[derive(Clone, Debug, Default)]
enum ListState {
  #[default]
  None,
  Unordered,
  Ordered {
    counter: u16,
    digits: u8,
  },
}

#[derive(Clone, Debug, Default)]
struct FormatState {
  style: Style,
  padding: String,
  list_state: ListState,
}

fn format_tree(
  tree: Rc<Node>,
  state: &mut FormatState,
  text: &mut Text<'static>,
  mut previous_sibling_is_block: bool,
) -> bool {
  use markup5ever_rcdom::NodeData::*;
  let mut state = match &tree.data {
    Document | Doctype { .. } | Comment { .. } | ProcessingInstruction { .. } => state.to_owned(),
    Text { contents } => {
      let s: String = contents.clone().into_inner().into();
      let s = s.replace('\n', ""); // Lines are insignificant in HTML
      if previous_sibling_is_block && !s.is_empty() {
        text.lines.push(Line {
          spans: vec![Span::styled(state.padding.to_owned(), state.style)],
          alignment: None,
        });
        previous_sibling_is_block = false;
      }
      text
        .lines
        .last_mut()
        .unwrap()
        .spans
        .push(Span::styled(s, state.style));
      state.to_owned()
    },
    Element {
      name: QualName {
        ns: ns!(html),
        local: name,
        ..
      },
      attrs,
      ..
    } => match name.as_ref() {
      "br" => {
        text.lines.push(Line::raw(state.padding.to_owned()));
        state.to_owned()
      },
      "p" => {
        previous_sibling_is_block = true;
        state.to_owned()
      },
      "blockquote" => {
        previous_sibling_is_block = true;
        FormatState {
          padding: state.padding.to_owned() + "> ",
          ..state.to_owned()
        }
      },
      "ul" => {
        previous_sibling_is_block = true;
        FormatState {
          list_state: ListState::Unordered,
          ..state.to_owned()
        }
      },
      "ol" => {
        previous_sibling_is_block = true;
        // FIXME: <li> are not guaranteed to be direct children, are they?
        let items_count: u16 = tree
          .children
          .borrow()
          .iter()
          .map(|child| match &child.data {
            Element {
              name:
                QualName {
                  ns: ns!(html),
                  local,
                  ..
                },
              ..
            } if local.as_ref() == "li" => 1,
            _ => 0,
          })
          .sum();
        FormatState {
          list_state: ListState::Ordered {
            counter: 0,
            digits: count_digits(items_count),
          },
          ..state.to_owned()
        }
      },
      "li" => {
        previous_sibling_is_block = false;
        match state.list_state {
          ListState::None => state.to_owned(),
          ListState::Unordered => {
            let mut line = Line::default();
            line
              .spans
              .push(Span::styled(state.padding.to_owned(), state.style));
            line.spans.push(Span::styled("* ", state.style));
            text.lines.push(line);
            FormatState {
              padding: state.padding.to_owned() + "  ",
              ..state.to_owned()
            }
          },
          ListState::Ordered {
            ref mut counter,
            digits,
          } => {
            let mut line = Line::default();
            line
              .spans
              .push(Span::styled(state.padding.to_owned(), state.style));
            *counter += 1;
            line
              .spans
              .push(Span::styled(format!("{}. ", counter), state.style));
            if count_digits(*counter) != digits {
              line.spans.push(Span::styled(
                " ".repeat((digits - count_digits(*counter)).into()),
                state.style,
              ));
            }
            text.lines.push(line);
            FormatState {
              padding: state.padding.to_owned() + "   " + &" ".repeat(digits.into()),
              ..state.to_owned()
            }
          },
        }
      },
      "em" | "i" => FormatState {
        style: state.style.italic(),
        ..state.to_owned()
      },
      "strong" | "b" => FormatState {
        style: state.style.bold(),
        ..state.to_owned()
      },
      "u" => FormatState {
        style: state.style.underlined(),
        ..state.to_owned()
      },
      _ => state.to_owned(),
    },
    Element { .. } => state.to_owned(), // Element not in the HTML namespace
  };

  for subtree in tree.children.borrow().iter() {
    previous_sibling_is_block =
      format_tree(subtree.clone(), &mut state, text, previous_sibling_is_block);
  }

  match &tree.data {
    Document | Doctype { .. } | Comment { .. } | ProcessingInstruction { .. } => {},
    Text { .. } => {},
    Element {
      name: QualName {
        ns: ns!(html),
        local: name,
        ..
      },
      attrs,
      ..
    } => match name.as_ref() {
      "p" | "blockquote" => {
        previous_sibling_is_block = true;
      },
      _ => {},
    },
    Element { .. } => {},
  }

  previous_sibling_is_block
}

pub fn format_html(s: &str) -> Text<'static> {
  let tree = parse_fragment(
    RcDom::default(),
    Default::default(),
    QualName::new(None, ns!(html), local_name!("body")),
    Vec::new(),
  )
  .one(s)
  .document;
  let mut state = FormatState {
    padding: "    ".to_owned(),
    ..Default::default()
  };
  let mut text = Text::raw("");
  format_tree(tree, &mut state, &mut text, false);
  text
}
